/*
 * This Kotlin source file was generated by the Gradle 'init' task.
 */
package at.tugraz.iaik.p2p.pol

import at.tugraz.iaik.p2p.commons.CommonCrypto
import at.tugraz.iaik.p2p.commons.CryptoException
import at.tugraz.iaik.p2p.evt.msg.Error
import at.tugraz.iaik.p2p.evt.msg.ErrorMessage
import at.tugraz.iaik.p2p.evt.msg.EvalRequestMessage
import at.tugraz.iaik.p2p.evt.msg.EvalResponseMessage
import at.tugraz.iaik.p2p.skad.config.Config
import at.tugraz.iaik.p2p.skad.id.ECID
import at.tugraz.iaik.p2p.skad.msg.ExtendedMessage
import at.tugraz.iaik.p2p.skad.msg.Message
import at.tugraz.iaik.p2p.skad.node.LocalNode
import com.codahale.shamir.Scheme
import kotlinx.coroutines.Deferred
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.async
import kotlinx.coroutines.awaitAll
import mu.KotlinLogging
import java.io.ByteArrayInputStream
import java.io.ByteArrayOutputStream
import java.io.DataInputStream
import java.io.DataOutputStream
import javax.crypto.Cipher
import javax.crypto.spec.SecretKeySpec

//P2P Node capable of issuing decrypt requests and handling incoming policy evaluaton requests
class PolNode(private val config: Config) {

    //kademlia node responsible for p2p communication with listener for custom messages
    private val node = LocalNode(config, incomingMsgListener = ::listen)
    val logger = KotlinLogging.logger("P ${node.id}")
    val identifier = node.id

    //
    suspend fun bootstrap(node: PolNode) = this.node.bootstrap(node.node)

    suspend fun decrypt(container: SignedDataContainer): ByteArray {


        val responses = mutableListOf<Deferred<Message>>()
        container.dataContainer.protectedData.dataProtection.shares.forEach {
            val decryptRequestMessage = EvalRequestMessage(
                CommonCrypto.RND.nextLong(),
                node,
                it,
                container.dataContainer.signedPolicy,
                container.dataContainer.creator
            )
            node.findNode(it.evaluator)?.let {
                GlobalScope.async {
                    node.actor.send(decryptRequestMessage, it)
                }.let { responses.add(it) }
            }
        }
        val decryptResponses = responses.awaitAll().map { it.apply { if (it is ErrorMessage) logger.warn { it } } }
            .filterIsInstance<EvalResponseMessage>()


        if (decryptResponses.size < container.dataContainer.protectedData.dataProtection.k.toInt())
            throw NotEnoughSharesException(
                container.dataContainer.protectedData.dataProtection.k.toInt(),
                decryptResponses.size
            )

        val scheme = Scheme(
            CommonCrypto.RND,
            container.dataContainer.protectedData.dataProtection.shares.size,
            container.dataContainer.protectedData.dataProtection.k.toInt()
        )
        val recoveryInputs = mutableMapOf<Int, Share>()
        container.dataContainer.protectedData.dataProtection.shares.forEachIndexed { index, share ->
            decryptResponses.firstOrNull { it.origin.id == share.evaluator }
                ?.let {

                    val decryptedShare = CommonCrypto.decrypt(node.config.privateKey!!, it.recoveredShare)

                    recoveryInputs[index + 1] = DataInputStream(ByteArrayInputStream(decryptedShare)).readShare()

                }
        }

        val validatedMap = mutableMapOf<Int, KeyShare>()

        recoveryInputs.forEach { (index, decryptedShare) ->
            val share = container.dataContainer.protectedData.dataProtection.shares[index - 1]

            if (share.evaluator != decryptedShare.evaluator) {
                logger.warn { "received share ${index - 1} evaluator mismatch: expected: ${share.evaluator}, actual: ${decryptedShare.evaluator}" }

            } else {
                try {
                    //check whether the decrypted share originated from the evaluator it should
                    CommonCrypto.validate(decryptedShare.fragment, decryptedShare.signature, share.evaluator.publicKey)
                    try {
                        //check whether the obtained key was actually created by the creator of the policy
                        CommonCrypto.validate(
                            decryptedShare.fragment,
                            share.signature,
                            container.dataContainer.creator.publicKey
                        )
                        validatedMap[index] = decryptedShare.fragment
                    } catch (t: Throwable) {
                        logger.warn { t }
                        t.printStackTrace()
                    }
                } catch (t: Throwable) {
                    logger.warn { t }
                    t.printStackTrace()
                }

            }
        }
        val secret = scheme.join(validatedMap)
        val cipher = Cipher.getInstance("AES/GCM/NoPadding")
        val key = SecretKeySpec(secret, "AES")
        cipher.init(Cipher.DECRYPT_MODE, key, container.dataContainer.protectedData.dataProtection.params)
        return cipher.doFinal(container.dataContainer.protectedData.ciphertext)
    }

    private suspend fun listen(msg: ExtendedMessage): ExtendedMessage {
        return when (msg) {
            is EvalRequestMessage -> {
                if (msg.share.evaluator != node.id)
                    return ErrorMessage(
                        msg.seqId,
                        node,
                        Error.EVL_MISMATCH,
                        "I am not responsible for share directed at ${msg.share.evaluator}"
                    )
                else {
                    val policy = msg.policy.policy
                    try {
                        CommonCrypto.validate(
                            policy.toByteArray(),
                            msg.policy.signature,
                            msg.creator.publicKey
                        )
                    } catch (t: CryptoException) {
                        return ErrorMessage(
                            msg.seqId,
                            node,
                            Error.INVALID_POLICY_SIG,
                            "Policy signature does not match creator id"
                        )
                    }

                    policy.setSender(msg.origin.id as ECID)
                    try {
                        policy.evaluate()
                    } catch (t: Throwable) {
                        return ErrorMessage(
                            msg.seqId,
                            node,
                            Error.UNAUTHORIZED,
                            t.message ?: ""
                        )
                    }


                    val recovered = CommonCrypto.decrypt(config.privateKey!!, msg.share.fragment)
                    try {
                        CommonCrypto.validate(recovered, msg.share.signature, msg.creator.publicKey)
                    } catch (t: CryptoException) {
                        return ErrorMessage(
                            msg.seqId,
                            node,
                            Error.INVALID_SHARE_SIG,
                            "key share signature does not match creator id"
                        )
                    }
                    val recoveredShare =
                        Share(node.id as ECID, recovered, CommonCrypto.sign(node.config.privateKey!!, recovered))
                    val baos = ByteArrayOutputStream(4096)
                    val buf = DataOutputStream(baos)
                    buf.writeShare(recoveredShare)

                    val arr = baos.toByteArray()
                    return EvalResponseMessage(
                        msg.seqId,
                        node,
                        CommonCrypto.encrypt((msg.origin.id as ECID).publicKey, arr)
                    )
                }
            }
            else -> ErrorMessage(msg.seqId, node, Error.UNKNOWN_MSG, "message not known!")
        }
    }

    class NotEnoughSharesException(val needed: Int, val present: Int, t: Throwable? = null) :
        Throwable("less than $needed shares received: $present", t)
}